"""A collection of functions and information for the MU system."""

import pathlib
from pathlib import Path
from typing import Any, Optional

import numpy as np
import scipy.interpolate

from pyant.models import Array, ArrayParams, InterpolatedArray, InterpolatedArrayParams
from radardef.radar_stations.mu.beams.data import DATA_PATHS
from radardef.tools.types import CarthesianCoordinates_3xN, NDArray_2, NDArray_2xN
from spacecoords.spherical import cart_to_sph


def mu_array_beam() -> tuple[Array, ArrayParams]:
    """A MU array beam"""

    assert "MU_antenna_pos.npy" in DATA_PATHS, "pos file missing!"
    _mu_antennas = np.load(DATA_PATHS["MU_antenna_pos.npy"])

    assert "mu_yagi_gain.npz" in DATA_PATHS, "gain file missing!"
    _mu_yagi = np.load(DATA_PATHS["mu_yagi_gain.npz"])

    az = _mu_yagi["az_deg"].reshape(-1, 721)
    az -= 180
    el = _mu_yagi["el_deg"].reshape(-1, 721)
    gain_dB = _mu_yagi["gain_dB"].reshape(-1, 721)
    gain_dB = gain_dB - np.max(_mu_yagi["gain_dB"])

    interp = scipy.interpolate.RegularGridInterpolator(
        (az[0, :], el[:, 0]),
        gain_dB.T,
        bounds_error=False,
    )

    def yagi(cart_coord: CarthesianCoordinates_3xN, polarization: NDArray_2) -> NDArray_2xN:
        sph = cart_to_sph(cart_coord, degrees=True)
        G = 10 ** (interp(sph[:2, :].T) / 10.0)
        return np.stack([G, G], axis=0)

    beam = Array(
        antennas=_mu_antennas,
        antenna_element=yagi,
    )
    params = ArrayParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        frequency=46.5e6,
        polarization=beam.polarization.copy(),
    )
    return beam, params


def mu_interpolated_array_beam(
    path: Optional[Path] = None, **interpolation_kwargs: Any
) -> tuple[InterpolatedArray, InterpolatedArrayParams]:
    """A MU interpolated array beam"""
    beam = InterpolatedArray()
    params = InterpolatedArrayParams(pointing=np.array([0, 0, 1], dtype=np.float64))
    if path is None:
        array, arr_params = mu_array_beam()
        beam.generate_interpolation(array, arr_params, **interpolation_kwargs)
        return beam, params

    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)

    if path.is_file():
        beam.load(path)
    else:
        array, arr_params = mu_array_beam()
        beam.generate_interpolation(array, arr_params, **interpolation_kwargs)
        beam.save(path)
    return beam, params



"Collect the paths to the available data in a easy to access format"

import importlib.resources
import pathlib

DATA_PATHS = {}

# To be compatible with 3.7-8
# as resources.files was introduced in 3.9
if hasattr(importlib.resources, "files"):
    _data_files = importlib.resources.files(__name__)
    for file in _data_files.iterdir():
        if not file.is_file():
            continue
        if file.name.endswith(".py"):
            continue

        DATA_PATHS[file.name] = pathlib.Path(str(file))

else:
    _data_folder = importlib.resources.contents(__name__)
    for fname in _data_folder:
        with importlib.resources.path(__name__, fname) as file:
            if not file.is_file():
                continue
            if file.name.endswith(".py"):
                continue

            DATA_PATHS[file.name] = pathlib.Path(str(file))



"""Template class for all validators to inherit from"""

from abc import abstractmethod
from pathlib import Path
from typing import Generic, TypeVar

T = TypeVar("T")


class Validator(Generic[T]):
    """
    Validator template, should be inherited by all validators

    Args:
        format: Format that the validator should be validator for
    """

    @property
    def format(self) -> T:
        """Format connected to the validator"""
        return self.__format

    def __init__(self, format: T) -> None:
        self.__format = format

    def __str__(self) -> str:
        """validator format string"""
        return f"{self.__format} validator"

    @abstractmethod
    def validate(self, src: str | Path) -> bool:
        """
        Abstract method, validate the format of given file

        Args:
            src: Path to file/directory
        """
        pass



"""Template class for all data loaders to inherit from"""

from abc import abstractmethod
from pathlib import Path
from typing import Optional

import numpy as np
import numpy.typing as npt

from radardef.components import Validator
from radardef.types import Metadata, Pointing, TargetFormat


class DataLoader:
    """
    Data loader template, should be inherited by all data loaders

    Args:
        converted_format: Format of the converted data
        validator: validator for the converted format to see that a file is compatible

    """

    @property
    def converted_format(self) -> TargetFormat:
        """Converted format of the data compatible with the data loader"""
        return self.__converted_format

    @property
    @abstractmethod
    def meta(self) -> Metadata:
        """Metadata, containing experiment data and bounds data"""
        pass

    # TODO: change to carthesian coordinates
    @property
    @abstractmethod
    def pointing(self) -> Pointing:
        """Pointing data, data describing the radar pointing direction in spherical coordinates"""
        pass

    @property
    @abstractmethod
    def channels(self) -> list[int] | list[str]:
        """All available channels"""
        pass

    def __init__(self, converted_format: TargetFormat, validator: Validator[TargetFormat]):
        self.__converted_format = converted_format
        self.__validator = validator

    @abstractmethod
    def load(self, path: Path | str) -> None:
        """Loads a path to the dataloader, extracting metadata and other important specifications"""
        pass

    def validate(self, path: Path) -> bool:
        """Validate that the file format compatible with the loader"""
        return self.__validator.validate(path)

    @abstractmethod
    def bounds(self, channel: str | int) -> tuple[int, int]:
        """Sample bounds of the specific channel"""
        pass

    @abstractmethod
    def read(
        self,
        channel: str | int,
        start_sample: Optional[int] = None,
        vector_length: Optional[int] = None,
    ) -> npt.NDArray[np.complex128]:
        """

        Args:
            channel: channel to read data from
            start_sample (optional): sample to start reading from, if not given bounds start will be used
            vector_length (optional): Amount of samples to read from start_sample,
                if not given all samples will be read

        Returns:
            Complex data of given channel
        """
        pass



__all__ = ["Converter", "Validator", "RadarStation", "DataLoader"]
from radardef.components.converter_template import Converter
from radardef.components.validator_template import Validator

from radardef.components.radar_station_template import RadarStation  # isort: skip
from radardef.components.data_loader_template import DataLoader  # isort: skip



import configparser
import importlib.resources
import pathlib

import numpy as np
from numpy.typing import NDArray

EXP_FILES: dict[str, pathlib.Path] = {}

# To be compatible with 3.7-8
# as resources.files was introduced in 3.9
if hasattr(importlib.resources, "files"):
    _data_files = importlib.resources.files("radardef.radar_stations.eiscat.utils")
    for file in _data_files.iterdir():
        if isinstance(file, pathlib.Path):
            if not file.is_file():
                continue
            if file.name.endswith(".py"):
                continue

            EXP_FILES[file.name] = file.resolve()

else:
    _data_content = importlib.resources.contents("radardef.radar_stations.eiscat.utils")
    for fname in _data_content:
        with importlib.resources.path("radardef.radar_stations.eiscat.utils", fname) as file:
            if not file.is_file():
                continue
            if file.name.endswith(".py"):
                continue

            EXP_FILES[file.name] = pathlib.Path(str(file)).resolve()


def load_radar_code(xpname: str) -> NDArray[np.float64]:
    """Load radar code, xpname + _code.txt."""

    code_name = xpname + "_code.txt"
    assert code_name in EXP_FILES, f"radar code '{code_name}' not found in pre-defined configurations"
    code_file = EXP_FILES[code_name]
    try:
        with open(code_file, "r") as fh:
            code = []
            for line in fh:
                code.append([1 if ch == "+" else -1 for ch in line.strip()])
        return np.array(code, dtype=np.float64)
    except Exception as e:
        raise ValueError(f"Couldn't open code file for {xpname}:" + str(e))


def load_expconfig(xpname: str) -> configparser.ConfigParser:
    """load expconfig, any .ini file"""
    cfg_name = xpname + ".ini"
    assert cfg_name in EXP_FILES, f'experiment "{cfg_name}" not found in pre-defined configurations'
    cfg_file = EXP_FILES[cfg_name]
    try:
        cfg = configparser.ConfigParser()
        cfg.read_file(open(cfg_file, "r"))
        return cfg
    except Exception as e:
        raise ValueError(f"Couldn't open config file for {xpname}:" + str(e))



"""A collection of functions and information for the EISCAT UHF Radar system."""

import numpy as np

from pyant.beam import Beam
from pyant.models import (
    Cassegrain,
    CassegrainParams,
    MeasuredAzimuthallySymmetric,
    MeasuredAzimuthallySymmetricParams,
)
from pyant.models.measured import InterpMethods

from .data import DATA_PATHS


def eiscat_uhf_beam(
    interpolation_method: InterpMethods,
) -> tuple[MeasuredAzimuthallySymmetric, MeasuredAzimuthallySymmetricParams]:
    """Eiscat uhf measured azimuthally symmetric beam"""

    assert "eiscat_uhf_bp.txt" in DATA_PATHS, "data file missing!"

    with open(DATA_PATHS["eiscat_uhf_bp.txt"], "r") as stream:
        eiscat_beam_data = np.genfromtxt(stream)

    peak_gain = 10**4.81
    beam = MeasuredAzimuthallySymmetric(
        off_axis_angle=eiscat_beam_data[:, 0],
        gains=10.0 ** (eiscat_beam_data[:, 1] / 10.0) * peak_gain,
        interpolation_method=interpolation_method,
        degrees=True,
    )
    params = MeasuredAzimuthallySymmetricParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
    )
    return beam, params


def eiscat_uhf_cassegrain_beam() -> tuple[Cassegrain, CassegrainParams]:
    """Eiscat uhf cassegrain beam"""

    beam = Cassegrain(
        peak_gain=10**4.81,
    )
    params = CassegrainParams(
        pointing=np.array([0, 0, 1], dtype=np.float64),
        frequency=930e6,
        outer_radius=40.0,
        inner_radius=23.0,
    )
    return beam, params



import setuptools

setuptools.setup(
    packages=setuptools.find_packages(where="src"),
    package_dir={"": "src"},
    package_data={"": ["*.ini", "*.txt", "*.npz", "*.npy", "antpos.csv*"]},
    include_package_data=True,
)



"""
The CLI format functionality, abstracts the format functionality to a user friendly CLI interface
"""

import argparse
import logging

from radardef.collections import FormatCollection
from radardef.radar_stations import EiscatUHF, Mu

from .commands_cli import add_command


def main(args: argparse.Namespace, cli_logger: logging.Logger) -> None:
    """Converter CLI available to the user"""

    format_validator = FormatCollection([Mu(), EiscatUHF()])

    if args.list:
        print(format_validator.list_formats())
        if cli_logger:
            cli_logger.info("Listing complete, exiting...")
        return

    for path in args.paths:
        source_format = format_validator.get_format(path)
        print(f"Source format is: {source_format}")


def parser_build(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Adds mandatory and optional positional arguments to the parser."""

    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="List available formats",
    )
    parser.add_argument(
        "paths",
        nargs="*",
        default=[],
        help="Location of the file/files",
    )

    return parser


add_command(
    name="format",
    function=main,
    parser_build=parser_build,
    command_help="Returns the source format of the file",
)



"""H5 validator"""

from pathlib import Path

from radardef.components.validator_template import Validator
from radardef.types import TargetFormat


class H5(Validator):
    """H5 validator"""

    def __init__(self) -> None:
        super().__init__(TargetFormat.H5)

    def validate(self, src: str | Path) -> bool:
        """Validate that the path is a h5 file of the correct format"""
        path = Path(src).resolve()

        if path.is_file():
            return self._is_h5_file(path)
        else:
            files = [f for f in path.iterdir() if self._is_h5_file(f)]
            return len(files) > 0

    def _is_h5_file(self, src: Path) -> bool:
        """Compatible h5 file"""
        return (len(src.name) == 32) and (src.name[10] == "T") and (src.suffix == ".h5")



"""
The CLI convert functionality, abstracts the convert functionality to a user friendly CLI interface
"""

import argparse
import logging
import os
from pathlib import Path

from radardef import RadarDef
from radardef.types import TargetFormat

from .commands_cli import add_command


def main(args: argparse.Namespace, cli_logger: logging.Logger) -> None:
    """Converter CLI available to the user"""
    radar_def = RadarDef()

    if args.list:
        print(radar_def.converter_collection.list_collection())
        if cli_logger:
            cli_logger.info("Listing complete, exiting...")
        return

    # If no files available, exit
    if len(args.files) == 0:
        print("No input paths!")
        return

    target_format = args.format.lower()

    # Extract wanted target format
    if target_format not in iter(TargetFormat):
        target_format = TargetFormat.UNKNOWN
        print("No conversion to target format is present, use -l to list available formats")
        return

    converted_data = radar_def.convert(args.files, target_format, args.output)

    if converted_data is None:
        print("Path/paths was not valid")
        return

    if cli_logger:
        cli_logger.info("Conversion complete")


def parser_build(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Adds mandatory and optional positional arguments to the parser."""

    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="List avalible convertable formats and target format",
    )
    parser.add_argument(
        "files",
        nargs="*",
        default=[],
        help="Input the locations of the files (or folders) to be converted",
    )
    parser.add_argument(
        "format",
        default=None,
        help="Target format to convert to",
    )
    parser.add_argument(
        "-o",
        "--output",
        default=str(Path(os.getcwd()).resolve()),
        help="The output location of the converted files",
    )
    parser.add_argument(
        "-s",
        "--sourceformat",
        default=None,
        help="optional, format of source data, if specified all other formats will be ignored",
    )

    return parser


add_command(
    name="convert",
    function=main,
    parser_build=parser_build,
    command_help="Convert the target files to a supported format",
)



"""
ConverterCollection gathers all given radars and extracts their converters, the collection can then convert
between any given source format to a target format given that atleast one radar contained such a converter.
"""

import logging
from pathlib import Path
from typing import Iterator

import radardef.tools as tools
from radardef.components.converter_template import Converter
from radardef.components.radar_station_template import RadarStation
from radardef.types import SourceFormat, TargetFormat

try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
except ImportError:

    class COMM_WORLD:
        rank = 0
        size = 1

    comm = COMM_WORLD()  # type: ignore[assignment]


class ConverterCollection:
    """
    Collection of all available converters from the available radar objects,
    provides a way to gather and provide converters for many different formats in one place.

    Args:
        radars: List of radars that the collections converters will be extracted from.

    """

    __logger = logging.getLogger(__name__)

    def __init__(self, radars: list[RadarStation]):
        self.__converters: dict[SourceFormat, dict] = dict()

        for radar in radars:
            self._register_converters(radar.get_converters())

    def __iter__(self) -> Iterator:
        """Iterate over all available converters."""
        return iter(self.__converters)

    def _register_converters(self, converters: list[Converter]) -> None:
        """Register several converters to the collection."""
        for converter in converters:
            self.__logger.info(f"Registering {converter}")

            if converter.source_format in self.__converters:
                self.__converters[converter.source_format][converter.target_format] = converter
            else:
                self.__converters[converter.source_format] = {converter.target_format: converter}

    @tools.MPI_target_args(range(1, 3), loading_bar=True, MPI=comm.size > 1)
    def convert(
        self, path: Path, source_format: SourceFormat, target_format: TargetFormat, output_dir: Path
    ) -> list[Path] | None:
        """
        Converts the file/directory at the given path from a source format to a target format,
        parrellized with MPI if needed.
        For parallelization path list and source_format list has to be of the same size.
        """

        try:
            return self.__converters[source_format][target_format].convert(path, output_dir)
        except KeyError:
            self.__logger.info(
                f'No converter from source format: "{source_format} \
                to target format: {target_format}, files affected: {path}'
            )
        return None

    def list_collection(self) -> str:
        """Lists available converters as a string."""

        st = ""
        for source_format in self.__converters:
            st += f"{source_format}:\n"
            for target_format in self.__converters[source_format]:
                st += f"├Target format> {target_format}\n"
        return st

    def get_converter(self, source_format: SourceFormat, target_format: TargetFormat) -> Converter:
        return self.__converters[source_format][target_format]



"""MU radar object"""

from pyant import Beam
from pyant.types import Parameters
from radardef.components.radar_station_template import RadarStation
from radardef.types import StationID

from .beams import mu_array_beam, mu_interpolated_array_beam
from .converters import MuiToH5
from .data_loaders import H5Loader
from .validators import MUI


class Mu(RadarStation):
    """
    MU radar definition

    Args:
        interpolated (optional): If beam should be interpolated, by default False.

    """

    def __init__(self, interpolated: bool = False) -> None:

        beam: Beam
        params: Parameters
        if interpolated:
            beam, params = mu_interpolated_array_beam()
        else:
            beam, params = mu_array_beam()

        super().__init__(
            StationID.MU,
            transmitter=True,
            receiver=True,
            lat=34.85402777777778,
            lon=136.10562222222222,
            alt=372.0,
            beam=beam,
            beam_parameters=params,
            min_elevation=30,
            power=1e6,
            power_per_element=2105.2631578947367,
            frequency=46.5e6,
            converters=[MuiToH5()],
            validator=MUI(),
            data_loaders=[H5Loader()],
        )



__all__ = ["ConverterCollection", "DataLoaderCollection", "FormatCollection"]
from .converter_collection import ConverterCollection
from .data_loader_collection import DataLoaderCollection
from .format_collection import FormatCollection



"""
This module contains convenient type information so that typing can be precise
but not too verbose in the code itself.
"""

import numpy as np
import numpy.typing as npt

NDArray_N = npt.NDArray
"(n,) shaped ndarray"

NDArray_2 = npt.NDArray
"(2,) shaped ndarray"

NDArray_3 = npt.NDArray
"(3,) shaped ndarray"

NDArray_3x1 = npt.NDArray
"(3,) shaped ndarray"

NDArray_2xN = npt.NDArray
"(2,n) shaped ndarray"

NDArray_3xN = npt.NDArray
"(3,n) shaped ndarray"

NDArray_3xNxM = npt.NDArray
"(3,n,m) shaped ndarray"

NDArray_MxM = npt.NDArray
"(m,m) shaped ndarray"

CarthesianCoordinates_3xN = npt.NDArray[np.float64]
""" CarthesianCoordinates [x,y,z]"""



"""Eiscat UHF radar object"""

from pyant.beam import Beam
from pyant.models.measured import InterpMethods
from pyant.types import Parameters
from radardef.components.radar_station_template import RadarStation
from radardef.radar_stations.eiscat.beams.uhf import (
    eiscat_uhf_beam,
    eiscat_uhf_cassegrain_beam,
)
from radardef.types import BeamType, EiscatUHFLocation, StationID

from .converters import EiscatMatbzToDrf
from .data_loaders import DrfLoader
from .validators import EiscatMatlab


class EiscatUHF(RadarStation):
    """
    Eiscat UHF radar definition

    Args:
        location: Eiscat uhf radar location (tromso, kiruna, sodankyla)
        beam_type (optional): Beam type (measured, cassegrain)
        interpolation_method: What interpolation method to use

    """

    def __init__(
        self,
        location: EiscatUHFLocation = EiscatUHFLocation.KIRUNA,
        beam_type: BeamType = BeamType.MEASURED,
        interpolation_method: InterpMethods = "linear",
    ) -> None:

        beam: Beam
        params: Parameters
        if beam_type == BeamType.MEASURED:
            beam, params = eiscat_uhf_beam(interpolation_method)
        elif beam_type == BeamType.CASSEGRAIN:
            beam, params = eiscat_uhf_cassegrain_beam()

        match location:
            case EiscatUHFLocation.KIRUNA:
                super().__init__(
                    StationID.EISCAT_UHF_KIRUNA,
                    transmitter=False,
                    receiver=True,
                    lat=67.86055555555555,
                    lon=20.435277777777777,
                    alt=418.0,
                    beam=beam,
                    beam_parameters=params,
                    min_elevation=30,
                    noise_temperature=100,
                    frequency=930.0e6,
                    converters=[EiscatMatbzToDrf()],
                    validator=EiscatMatlab(),
                    data_loaders=[DrfLoader()],
                )
            case EiscatUHFLocation.SODANKYLA:
                super().__init__(
                    StationID.EISCAT_UHF_SODANKYLA,
                    transmitter=False,
                    receiver=True,
                    lat=67.36361111111111,
                    lon=26.626944444444444,
                    alt=197.0,
                    beam=beam,
                    beam_parameters=params,
                    min_elevation=30,
                    noise_temperature=100,
                    frequency=930.0e6,
                    converters=[EiscatMatbzToDrf()],
                    validator=EiscatMatlab(),
                    data_loaders=[DrfLoader()],
                )
            case EiscatUHFLocation.TROMSO:
                super().__init__(
                    StationID.EISCAT_UHF_TROMSO,
                    transmitter=True,
                    receiver=True,
                    lat=69.58638888888889,
                    lon=19.227222222222224,
                    alt=86.0,
                    beam=beam,
                    beam_parameters=params,
                    min_elevation=30,
                    noise_temperature=100,
                    power=1.6e6,
                    frequency=930.0e6,
                    converters=[EiscatMatbzToDrf()],
                    validator=EiscatMatlab(),
                    data_loaders=[DrfLoader()],
                )
